"""
The go starter template.

Author: Tom Fleet
Created: 24/06/2021
"""


import shutil
import subprocess
from pathlib import Path
from typing import List, Optional

from pytoil.exceptions import GoNotInstalledError
from pytoil.starters.base import BaseStarter


class GoStarter(BaseStarter):
    """
    The go starter template class.
    """

    def __init__(self, path: Path, name: str) -> None:
        """
        The pytoil go starter template.

        Args:
            path (Path): Root path under which to generate the
                project from this template.
            name (str): The name of the project to be created.
        """
        self._path = path
        self._name = name
        self._files = ["README.md", "main.go"]

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(path={self.path!r}, name={self.name!r})"

    @property
    def path(self) -> Path:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def root(self) -> Path:
        return self._path.joinpath(self._name)

    @property
    def files(self) -> List[Path]:
        return [self.root.joinpath(filename) for filename in self._files]

    def raise_for_go(self) -> None:
        """
        Raises an error if the user doesn't have go installed.
        """
        if not bool(shutil.which("go")):
            raise GoNotInstalledError("Go not found on $PATH.")

    def generate(self, username: Optional[str] = None) -> None:
        """
        Generate a new go starter template.

        This is a mix of creating files in python, and invoking
        `go mod init` in a subprocess to initialise the go
        modules file.
        """

        # Must have go installed to run go mod init
        self.raise_for_go()

        # Make the parent directory
        self.root.mkdir(parents=True)

        for file in self.files:
            file.touch()

        # Put the header in the readme
        readme = self.root.joinpath("README.md")
        readme.write_text(f"# {self.name}\n", encoding="utf-8")

        # Populate the main.go file
        go_file = self.root.joinpath("main.go")
        go_text = 'package main\n\nimport "fmt"\n\nfunc main() {\n\tfmt.Println("Hello World")\n}\n'  # noqa: E501

        go_file.write_text(go_text, encoding="utf-8")

        # Invoke go mod init
        _ = subprocess.run(
            ["go", "mod", "init", f"github.com/{username}/{self.name}"],
            check=True,
            cwd=self.root,
            capture_output=True,
        )



"""
The `pytoil remove` subcommand group.

Author: Tom Fleet
Created: 18/06/2021
"""

import shutil
from typing import List

import typer
from wasabi import msg

from pytoil.cli import utils
from pytoil.config import Config

app = typer.Typer()


@app.command()
def remove(
    projects: List[str] = typer.Argument(
        None,
        help="Name of the project(s) to delete.",
        show_default=False,
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force delete without confirmation.",
        show_default=False,
    ),
    all_: bool = typer.Option(
        False,
        "--all",
        "-a",
        help="Delete all of your local projects.",
        show_default=False,
    ),
) -> None:
    """
    Remove projects from your local filesystem.

    The remove command provides an easy interface for decluttering your local
    projects directory.

    You can selectively remove any number of projects by passing them as
    arguments or nuke the whole lot with "--all/-a" if you want.

    As with most programmatic deleting, the directories are deleted instantly and
    not sent to trash. As such, pytoil will prompt you for confirmation before
    doing anything.

    The "--force/-f" flag can be used to force deletion without the confirmation
    prompt. Use with caution!

    Examples:

    $ pytoil remove project1 project2 project3

    $ pytoil remove project1 project2 project3 --force

    $ pytoil remove --all

    $ pytoil remove --all --force
    """
    config = Config.from_file()

    local_projects = utils.get_local_projects(path=config.projects_dir)

    if not local_projects:
        msg.warn("You don't have any local projects to remove!", spaced=True, exits=1)

    if not projects and not all_:
        msg.warn(
            "If not using the '--all' flag, you must specify projects to delete.",
            exits=1,
            spaced=True,
        )

    # If user gives a project that isn't here (e.g. typo), abort
    for project in projects:
        if project not in local_projects:
            msg.warn(
                f"{project!r} not found in projects directory. Was it a typo?",
                spaced=True,
                exits=1,
            )

    to_delete = local_projects if all_ else projects

    if not force:
        if all_:
            typer.confirm(
                "This will delete ALL of your projects. Are you sure?", abort=True
            )
            # We want to delete everything
            to_delete = local_projects

        elif len(projects) <= 5:
            typer.confirm(
                f"This will delete {', '.join(projects)} from your local filesystem. Are you sure?",  # noqa: E501
                abort=True,
            )
        else:
            # Too many to print out nicely
            typer.confirm(
                f"This will delete {len(projects)} projects from your local filesystem. Are you sure?",  # noqa: E501
                abort=True,
            )

    # If we get here, user has used --force or said yes when prompted

    for project in sorted(to_delete, key=str.casefold):
        typer.secho(f"Removing project: {project!r}.", fg=typer.colors.YELLOW)
        shutil.rmtree(config.projects_dir.joinpath(project))

    msg.good("Done!", spaced=True)



"""
The rust starter template.

Author: Tom Fleet
Created: 24/06/2021
"""

import shutil
import subprocess
from pathlib import Path
from typing import List, Optional

from pytoil.exceptions import CargoNotInstalledError
from pytoil.starters.base import BaseStarter


class RustStarter(BaseStarter):
    """
    Rust starter template class.
    """

    def __init__(self, path: Path, name: str) -> None:
        """
        The pytoil rust starter template.

        Args:
            path (Path): Root path under which to generate the
                project from this template.
            name (str): The name of the project to be created.
        """
        self._path = path
        self._name = name
        self._files = ["README.md"]

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(path={self.path!r}, name={self.name!r})"

    @property
    def path(self) -> Path:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def root(self) -> Path:
        return self._path.joinpath(self._name)

    @property
    def files(self) -> List[Path]:
        return [self.root.joinpath(filename) for filename in self._files]

    def raise_for_cargo(self) -> None:
        """
        Raises an error if user doesn't have cargo installed.
        """
        if not bool(shutil.which("cargo")):
            raise CargoNotInstalledError("Cargo not found on $PATH.")

    def generate(self, username: Optional[str] = None) -> None:
        """
        Generate a new cargo/rust starter template.

        This is basically all cargo, the only additional
        thing we do here is add in a README.
        """

        # Must have cargo installed
        self.raise_for_cargo()

        # Make the root dir
        self.root.mkdir(parents=True)

        # Invoke cargo in a subprocess to generate the project
        _ = subprocess.run(
            ["cargo", "init"],
            check=True,
            capture_output=True,
            cwd=self.root,
        )

        # Add in the readme
        for file in self.files:
            file.touch()

        readme = self.root.joinpath("README.md")
        readme.write_text(f"# {self.name}\n", encoding="utf-8")



"""
Global defaults for pytoil.

Author: Tom Fleet
Created: 18/06/2021
"""


import os
from pathlib import Path
from typing import List, Set

# Default path for pytoil's config file
CONFIG_FILE: Path = Path.home().joinpath(".pytoil.yml").resolve()

# Valid pytoil config keys
CONFIG_KEYS: Set[str] = {
    "projects_dir",
    "token",
    "username",
    "vscode",
    "common_packages",
    "init_on_new",
}

# Defaults for pytoil config
PROJECTS_DIR: Path = Path.home().joinpath("Development").resolve()
TOKEN: str = os.getenv("GITHUB_TOKEN") or ""
USERNAME: str = ""
VSCODE: bool = False
COMMON_PACKAGES: List[str] = []
INIT_ON_NEW: bool = True

# Config Schema
CONFIG_SCHEMA = """

- projects_dir (bool)

    The absolute path to where you keep your development projects
    (e.g. /Users/you/Projects)

- token (str)

    Your GitHub personal access token. This must have a minimum of repo
    read access. See the documentation here:
    https://docs.github.com/en/github/authenticating-to-github/keeping-your-account-and-data-secure/creating-a-personal-access-token

    Pytoil will try and get this from the config file initially, then fall back to the $GITHUB_TOKEN environment
    variable. If neither of these places are set, you will not be able to use pytoil commands that rely on the
    GitHub API. Pytoil will notify you of this when any of these commands are called.

- username (str)

    Your GitHub username.

- vscode (bool)

    Whether you want pytoil to open projects up using VSCode. This will happen on 'new' and 'checkout'.

- common_packages (List[str])

    A list of python package names to inject into every virtual environment pytoil creates
    (e.g. linters, formatters and other dev dependencies).

- init_on_new (bool)

    Whether or not you want pytoil to create an empty git repo when you make a new project with
    'pytoil new'. This can also be disabled on a per use basis using the '--no-git' flag.
"""



"""
Module responsible for interacting with git via
subprocesses.

Author: Tom Fleet
Created: 19/06/2021
"""

import shutil
import subprocess
from pathlib import Path
from typing import Optional

from pytoil.config import defaults
from pytoil.exceptions import GitNotInstalledError


class Git:
    """
    Wrapper around the user installed `git`.
    """

    def __init__(self, git: Optional[str] = shutil.which("git")) -> None:
        """
        A handy wrapper around the user installed `git`.

        Args:
            git (Optional[str], optional): The path to the `git` binary
                primarily used for separation of concerns during testing.
                Defaults to shutil.which("git").
        """
        self.git = git

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(git={self.git!r})"

    def raise_for_git(self) -> None:
        """
        Raises an exception if the user does not have
        git installed.
        """
        if not self.git:
            raise GitNotInstalledError(
                "'git' executable not found on $PATH. Is git installed?"
            )

    def run(
        self,
        *args: str,
        check: bool = True,
        cwd: Path = defaults.PROJECTS_DIR,
        silent: bool = False,
    ) -> None:
        """
        Generic method to run `git` in a subprocess.

        Pass any args to git via *args.

        e.g. 'run("clone", "https://github.com/me/repo.git")'

        Args:
            check (bool, optional): Will raise a CalledProcessError if
                something goes wrong. Defaults to True.

            cwd (Path, optional): Working directory of the subprocess.
                Defaults to defaults.PROJECTS_DIR.

            silent (bool): Whether to suppress showing the clone output.
                Defaults to False.

        Raises:
            subprocess.CalledProcessError: If the subprocess command fails.
        """

        self.raise_for_git()

        try:
            subprocess.run(
                [f"{self.git}", *args], check=check, cwd=cwd, capture_output=silent
            )
        except subprocess.CalledProcessError:
            raise

    def clone(
        self,
        url: str,
        check: bool = True,
        cwd: Path = defaults.PROJECTS_DIR,
        silent: bool = True,
    ) -> None:
        """
        Convenience wrapper around `self.run` to clone a repo.

        Args:
            url (str): The clone url of the repo.

            check (bool, optional): Raise CalledProcessError on failure.
                Defaults to True.

            cwd (Path, optional): Working directory of the subprocess.
                Defaults to defaults.PROJECTS_DIR.

            silent (bool): Whether to suppress showing the clone output.
                Defaults to True.

        Raises:
            subprocess.CalledProcessError: If the subprocess command fails.
        """

        self.run("clone", url, check=check, cwd=cwd, silent=silent)

    def init(self, path: Path, check: bool = True, silent: bool = False) -> None:
        """
        Convenience wrapper around `self.run` to initialise a repo.

        Args:
            path (Path): Root of the project to initialise
                the repo in.

            check (bool, optional): Raise CalledProcessError on failure.
                Defaults to True.

            silent (bool): Whether to suppress showing the clone output.
                Defaults to False.

        Raises:
            subprocess.CalledProcessError: If the subprocess command fails.
        """

        self.run("init", check=check, cwd=path, silent=silent)

    def set_upstream(
        self,
        owner: str,
        repo: str,
        path: Path,
        check: bool = True,
        silent: bool = False,
    ) -> None:
        """
        Sets the upstream repo for a local repo, e.g. on a cloned fork.

        Note difference between origin and upstream, origin of a cloned fork
        would be user/forked_repo where as upstream would be
        original_user/forked_repo.

        Args:
            owner (str): Owner of the upstream repo.
            repo (str): Name of the upstream repo.
            path (Path): Root of the project where the local git
                repo is.
            check (bool, optional): Raise CalledProcessError on failure.
                Defaults to True.
            silent (bool): Whether to suppress showing the clone output.
            Defaults to True.
        """
        base_url = "https://github.com"
        constructed_upstream = f"{base_url}/{owner}/{repo}.git"
        self.run(
            "remote",
            "add",
            "upstream",
            constructed_upstream,
            check=check,
            cwd=path,
            silent=silent,
        )



"""
Module responsible for handling python virtual environments
in a project with a requirements.txt (or requirements_dev.txt)

Author: Tom Fleet
Created: 15/07/2021
"""

import subprocess
from pathlib import Path
from typing import List

from pytoil.environments.virtualenv import Venv


class ReqTxtEnv(Venv):
    """
    requirements.txt | requirements_dev.txt EnvManager class.
    """

    def __init__(self, project_path: Path) -> None:
        """
        Representation of a virtual environment in a python
        project with only a requirements.txt
        (or requirements_dev.txt).

        Because this is all standard python virtual environment stuff
        here we subclass from the `Venv` class so that all we have to
        tweak are the install methods.

        Args:
            project_path (Path): The root path of the current project.
        """
        super().__init__(project_path=project_path)

    @property
    def info_name(self) -> str:
        return "requirements file"

    def install_self(self) -> None:
        """
        Installs everything in the requirements file
        into the virtual environment.
        """
        # Before installing from a requirements file
        # ensure the virtual env exists
        if not self.exists():
            self.create()

        # Prefer requirements_dev if present because it will have everything
        if self.project_path.joinpath("requirements_dev.txt").exists():
            requirements_file = "requirements_dev.txt"
        elif self.project_path.joinpath(
            "requirements.txt"
        ).exists():  # pragma: no cover
            # Coverage isn't picking this up but it's definitely tested
            requirements_file = "requirements.txt"

        cmd: List[str] = [
            f"{self.executable}",
            "-m",
            "pip",
            "install",
            "-r",
            f"{requirements_file}",
            "--quiet",
        ]

        subprocess.run(cmd, check=True, cwd=self.project_path)



"""
Assorted helper functions for the pytoil CLI.

Author: Tom Fleet
Created: 19/06/2021
"""

from pathlib import Path
from typing import Set

import httpx
from wasabi import msg

from pytoil.config import Config


def get_local_projects(path: Path) -> Set[str]:
    """
    Returns all the projects (directories) under
    `path`.
    """

    return {f.name for f in path.iterdir() if f.is_dir() and not f.name.startswith(".")}


def warn_if_no_api_creds(config: Config) -> None:
    """
    Will print a helpful warning message and exit the program
    if username or token are not filled out in the config file.
    """

    if not config.can_use_api():
        msg.warn(
            "You must fill set your username and token to use API features!",
            spaced=True,
            exits=1,
        )


def handle_http_status_errors(error: httpx.HTTPStatusError) -> None:
    """
    Handles a variety of possible HTTP Status errors, print's nicer output
    to the user, and exits the program if necessary.

    Call this in an except block on CLI commands accessing the
    GitHub API.

    Args:
        error (httpx.HTTPStatusError): The error to be handled.
    """
    code = error.response.status_code

    if code == 401:
        msg.fail(
            title="HTTP Error: 401 - Unauthorized",
            text="This usually means something is wrong with your token!",
            exits=1,
        )
    elif code == 404:
        msg.fail(
            title="HTTP Error: 404 - Not Found",
            text="This usually means you've made a typo.",
            exits=1,
        )
    elif code == 500:
        msg.fail(
            title="HTTP Error: 500 - Internal Server Error",
            text="This usually means GitHub is not happy.",
            exits=1,
        )



"""
The python starter template.

Author: Tom Fleet
Created: 23/06/2021
"""

from pathlib import Path
from typing import List, Optional

from pytoil.starters.base import BaseStarter


class PythonStarter(BaseStarter):
    """
    Python starter template class.
    """

    def __init__(self, path: Path, name: str) -> None:
        """
        The pytoil python starter template.

        Args:
            path (Path): Root path under which to generate the
                project from this template.
            name (str): The name of the project to be created.
        """
        self._path = path
        self._name = name
        self._files: List[str] = ["README.md", "requirements.txt", f"{self.name}.py"]

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(path={self.path!r}, name={self.name!r})"

    @property
    def path(self) -> Path:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def root(self) -> Path:
        return self._path.joinpath(self._name)

    @property
    def files(self) -> List[Path]:
        return [self.root.joinpath(filename) for filename in self._files]

    def generate(self, username: Optional[str] = None) -> None:
        """
        Generate a new python starter template.

        Because this is a python starter, we can just
        create all the files directly.
        """
        # Make the parent directory
        self.root.mkdir(parents=True)

        for file in self.files:
            file.touch()

        # Put the header in the readme
        readme = self.root.joinpath("README.md")
        readme.write_text(f"# {self.name}\n", encoding="utf-8")

        # Put a hint in the requirements.txt
        reqs = self.root.joinpath("requirements.txt")
        reqs.write_text(
            "# Put your requirements here e.g. flask>=1.0.0\n", encoding="utf-8"
        )

        # Populate the python file
        py_file = self.root.joinpath(f"{self.name}.py")
        py_text = (
            'def hello(name: str = "world") -> None:\n    print(f"hello {name}")\n'
        )
        py_file.write_text(py_text, encoding="utf-8")



"""
The pytoil new command.

Author: Tom Fleet
Created: 25/06/2021
"""

from typing import List

import httpx
import typer
from cookiecutter.main import cookiecutter
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.environments import Conda, Environment, Venv, VirtualEnv
from pytoil.exceptions import (
    CargoNotInstalledError,
    EnvironmentAlreadyExistsError,
    GoNotInstalledError,
)
from pytoil.git import Git
from pytoil.repo import Repo
from pytoil.starters import GoStarter, PythonStarter, RustStarter, Starter
from pytoil.vscode import VSCode

app = typer.Typer()


@app.command(context_settings={"allow_extra_args": True})
def new(
    ctx: typer.Context,
    project: str = typer.Argument(
        ...,
        help="Name of the project to create.",
    ),
    cookie: str = typer.Option(
        None,
        "--cookie",
        "-c",
        help="URL to a cookiecutter template repo from which to build the project.",
    ),
    starter: Starter = typer.Option(
        Starter.none,
        "--starter",
        "-s",
        help="Use a language-specific starter template",
        case_sensitive=False,
        show_default=True,
    ),
    venv: VirtualEnv = typer.Option(
        VirtualEnv.none,
        "--venv",
        "-v",
        help="Which type of virtual environment to create for the project.",
        case_sensitive=False,
        show_default=True,
    ),
    no_git: bool = typer.Option(
        False,
        "--no-git",
        "-n",
        help="Don't initialise an empty git repo in the root of the project.",
    ),
) -> None:
    """
    Create a new development project.

    Bare usage will simply create an empty folder in your configured projects
    directory.

    You can also create a project from a cookiecutter template by passing a valid
    url to the '--cookie/-c' flag.

    If you just want a very simple, language-specific starting template, use the
    '--starter/-s' option.

    By default, pytoil will initialise an empty git repo in the folder, following
    the style of modern language build tools such as rust's cargo. You can disable
    this behaviour by setting 'init_on_new' to false in pytoil's config file
    or by passing the '--no-git/-n' flag here.

    If you want pytoil to create a new virtual environment for your project, you
    can use the '--venv/-v' flag. Standard python and conda virtual environments
    are supported.

    If the '--venv/-v' flag is used, you may also pass a list of python packages
    to install into the created virtual environment. These will be delegated to
    the appropriate tool (pip or conda) depending on what environment was created.
    If the environment is conda, the packages will be passed at environment creation
    time meaning they will have their dependencies resolved together. Normal python
    environments will first be created and then have specified packages installed.

    If 'common_packages' is specified in pytoil's config file, these will automatically
    be included in the environment.

    To specify versions of packages via the command line, you must enclose them
    in double quotes e.g. "flask>=1.0.0" not flask>=1.0.0 otherwise this will
    be interpreted by the shell as a command redirection.

    Examples:

    $ pytoil new my_project

    $ pytoil new my_project --cookie https://github.com/some/cookie.git

    $ pytoil new my_project --venv conda

    $ pytoil new my_project -c https://github.com/some/cookie.git -v conda --no-git

    $ pytoil new my_project -v venv requests "flask>=1.0.0"

    $ pytoil new my_project --starter python
    """
    # Get config and ensure user can access API
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    # Setup the objects required
    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )
    code = VSCode(root=repo.local_path)
    git = Git()

    if ctx.args:
        packages: List[str] = config.common_packages + ctx.args
    else:
        packages = config.common_packages

    # Resolve config vs flag for no-git
    # flag takes priority over config
    use_git: bool = config.init_on_new and not no_git

    # Check is project already exists and warn/exit if so
    pre_new_checks(repo=repo, api=api)

    # Cant use --venv with non-python starters
    if (
        starter.value != starter.none  # User specified starter
        and starter.value != starter.python  # The starter is not python
        and venv.value != venv.none  # And the user wants a virtual environment
    ):
        msg.warn(
            f"Can't create a venv for {starter.value} project!",
            spaced=True,
            exits=1,
        )

    # If we get here, all is well and we can create stuff!
    make_new_project(
        repo=repo,
        git=git,
        cookie=cookie,
        starter=starter,
        use_git=use_git,
        config=config,
    )

    if venv.value == venv.venv:
        env = create_virtualenv(repo=repo, packages=packages)

        if config.vscode:
            msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
            code.set_workspace_python(env.executable)
            code.open()

    elif venv.value == venv.conda:
        env = create_condaenv(repo=repo, packages=packages)

        if config.vscode:
            msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
            code.set_workspace_python(env.executable)
            code.open()

    else:
        # Only other allowed condition is none
        typer.secho(
            "Virtual environment not requested. Skipping environment creation.",
            fg=typer.colors.YELLOW,
        )

        if config.vscode:
            msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
            code.open()


def make_new_project(
    repo: Repo, git: Git, cookie: str, starter: Starter, use_git: bool, config: Config
) -> None:
    """
    Create a new development project either from a cookiecutter
    template or from scratch.
    """
    # Can't use starter and cookiecutter at the same time
    if starter.value != Starter.none and cookie:
        msg.warn(
            "'--cookie' and '--starter' are mutually exclusive.",
            exits=1,
        )

    if cookie:
        # We don't initialise a git repo for cookiecutters
        # some templates have hooks which do this, mine do!
        msg.info(f"Creating {repo.name!r} from cookiecutter: {cookie!r}.")
        cookiecutter(template=cookie, output_dir=config.projects_dir)

    elif starter == Starter.go:
        msg.info(f"Creating {repo.name!r} from starter: {starter.value!r}.")
        go_st = GoStarter(path=config.projects_dir, name=repo.name)

        try:
            go_st.generate(username=config.username)
        except GoNotInstalledError:
            msg.fail("Error: Go not installed.", spaced=True, exits=1)

        if use_git:
            git.init(path=repo.local_path, check=True)

    elif starter == Starter.python:
        msg.info(f"Creating {repo.name!r} from starter: {starter.value!r}.")
        py_st = PythonStarter(path=config.projects_dir, name=repo.name)
        py_st.generate()

        if use_git:
            git.init(path=repo.local_path, check=True)

    elif starter == Starter.rust:
        msg.info(f"Creating {repo.name!r} from starter: {starter.value!r}.")
        rs_st = RustStarter(path=config.projects_dir, name=repo.name)

        try:
            rs_st.generate()
        except CargoNotInstalledError:
            msg.fail("Error: Cargo not installed.", spaced=True, exits=1)

    else:
        msg.info(f"Creating {repo.name!r} at {repo.local_path}.")
        # Make an empty dir and git repo
        repo.local_path.mkdir(parents=True)

        if use_git:
            git.init(path=repo.local_path, check=True)


def pre_new_checks(repo: Repo, api: API) -> None:
    """
    Checks whether the repo already exists either locally
    or remotely, prints helpful warning messages and exits
    the program if True.
    """

    is_local = repo.exists_local()

    try:
        is_remote = repo.exists_remote(api=api)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        if is_local:
            msg.warn(
                title=f"{repo.name!r} already exists locally!",
                text=f"To checkout this project, use 'pytoil checkout {repo.name}'.",
                spaced=True,
                exits=1,
            )
        elif is_remote:
            msg.warn(
                title=f"{repo.name!r} already exists on GitHub!",
                text=f"To checkout this project, use 'pytoil checkout {repo.name}'.",
                spaced=True,
                exits=1,
            )


def create_virtualenv(repo: Repo, packages: List[str]) -> Environment:
    """
    Creates and returns new virtual environment with packages and reports
    to user.
    """

    msg.info(
        f"Creating virtual environment for {repo.name!r}",
        text=f"Including packages: {', '.join(packages)}",
        spaced=True,
    )
    env = Venv(project_path=repo.local_path)
    with msg.loading("Working..."):
        env.create(packages=packages)

    return env


def create_condaenv(repo: Repo, packages: List[str]) -> Environment:
    """
    Creates and returns new conda environment with packages and reports
    to user.
    """
    msg.info(
        f"Creating conda environment for {repo.name!r}",
        text=f"Including packages: {', '.join(packages)}",
        spaced=True,
    )
    env = Conda(name=repo.name, project_path=repo.local_path)
    try:
        with msg.loading("Working..."):
            env.create(packages=packages)
    except EnvironmentAlreadyExistsError:
        msg.warn(
            f"Conda environment {env.name!r} already exists!", spaced=True, exits=1
        )

    return env



"""
Exceptions implemented in pytoil.
"""


class PytoilException(Exception):
    """
    Base pytoil exception from which all subclasses
    must inherit.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class ExternalToolNotInstalledException(PytoilException):
    """
    Base exception for any child exception responsible
    for raising in the presence of a required external
    tool that's not installed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class GitNotInstalledError(ExternalToolNotInstalledException):
    """
    Raise when calling something that needs the user
    to have git installed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class CodeNotInstalledError(ExternalToolNotInstalledException):
    """
    Raise when calling something that needs the user to have
    VSCode (specifically the `code` command) installed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class MissingInterpreterError(PytoilException):
    """
    Trying to install or otherwise manipulate a python
    virtual environment that does not have a valid python
    interpreter.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class CondaNotInstalledError(ExternalToolNotInstalledException):
    """
    Trying to do something that requires the user to have
    the `conda` package manager installed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class EnvironmentAlreadyExistsError(PytoilException):
    """
    Trying to overwrite an existing environment, only applicable
    to conda environments.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class BadEnvironmentFileError(PytoilException):
    """
    The conda environment's `environment.yml` is malformed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class EnvironmentDoesNotExistError(PytoilException):
    """
    Trying to do something to a virtual environment that does not
    exist.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class UnsupportedCondaInstallationError(PytoilException):
    """
    User's conda installation is not one of the supported
    ones for pytoil. See environments/conda.py.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class RepoNotFoundError(PytoilException):
    """
    The repo object trying to be operated on does not exist.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class GoNotInstalledError(ExternalToolNotInstalledException):
    """
    The user does not have `go` installed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class CargoNotInstalledError(ExternalToolNotInstalledException):
    """
    The user does not have `cargo` installed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class FlitNotInstalledError(ExternalToolNotInstalledException):
    """
    The user does not have `flit` installed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class PoetryNotInstalledError(ExternalToolNotInstalledException):
    """
    The user does not have `poetry` installed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)



"""
Module responsible for creating and managing virtual environments
for flit.

This is done in a very similar way to `virtualenv.py` as flit
does not create it's own environments.

Author: Tom Fleet
Created: 13/07/2021
"""

import shutil
import subprocess
from pathlib import Path
from typing import List

from pytoil.environments.virtualenv import Venv
from pytoil.exceptions import FlitNotInstalledError


class FlitEnv(Venv):
    """
    Flit EnvManager class.
    """

    def __init__(self, project_path: Path) -> None:
        """
        Representation of a flit virtualenv.

        Actually not overtly different from the `Venv` class
        in `virtualenv.py` as flit uses standard python
        virtual environments.

        The main difference is found in install commands. Hence why
        we subclass from the standard `Venv` class here and other
        classes subclass from the `Environment` ABC.

        Args:
            project_path (Path): The root path of the current project.
        """
        super().__init__(project_path=project_path)

    @property
    def info_name(self) -> str:
        return "flit"

    def install_self(self) -> None:
        """
        Installs the current package.

        Raises:
            FlitNotInstalledError: If `flit` not installed.
        """

        if not bool(shutil.which("flit")):
            raise FlitNotInstalledError(
                "Flit not installed, cannot install flit based project."
            )

        # Unlike poetry, conda etc. flit does not make it's own virtual environment
        # we must make one here before installing the project
        if not self.exists():
            self.create()

        cmd: List[str] = [
            "flit",
            "install",
            "--deps",
            "develop",
            "--symlink",
            "--python",
            f"{self.executable}",
        ]

        # Here we must specify the cwd as flit will search for it's pyproject.toml
        subprocess.run(cmd, check=True, cwd=self.project_path, capture_output=True)



"""
The `pytoil pull` subcommand group.

Author: Tom Fleet
Created: 18/06/2021
"""

import concurrent.futures
from typing import List, Set

import httpx
import typer
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.git import Git
from pytoil.repo import Repo

app = typer.Typer()


@app.command()
def pull(
    projects: List[str] = typer.Argument(
        None,
        help="Name of the project(s) to pull down.",
        show_default=False,
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force pull without confirmation.",
        show_default=False,
    ),
    all_: bool = typer.Option(
        False,
        "--all",
        "-a",
        help="Pull down all of your projects.",
        show_default=False,
    ),
) -> None:
    """
    Pull down your remote projects.

    The pull command provides easy methods for pulling down remote projects.

    It is effectively a nice wrapper around git clone but you don't have to
    worry about urls or what your cwd is, pull will grab your remote projects
    by name and clone them to your configured projects directory.

    You can also use pull to batch clone multiple repos, even all of them ("--all/-a")
    if you're into that sorta thing.

    If more than 1 repo is passed (or if "--all/-a" is used) pytoil will pull
    the repos concurrently, speeding up the process.

    Any remote project that already exists locally will be skipped and none of
    your local projects are changed in any way. pytoil will only pull down
    those projects that don't already exist locally.

    It's very possible to accidentally clone a lot of repos when using pull so
    you will be prompted for confirmation before pytoil does anything.

    The "--force/-f" flag can be used to override this confirmation prompt if
    desired.

    Examples:

    $ pytoil pull project1 project2 project3

    $ pytoil pull project1 project2 project3 --force

    $ pytoil pull --all

    $ pytoil pull --all --force
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    if not projects and not all_:
        msg.warn(
            "If not using the '--all' flag, you must specify projects to pull.",
            exits=1,
            spaced=True,
        )

    try:
        not_found = False
        with msg.loading("Calculating difference..."):
            local_projects = utils.get_local_projects(path=config.projects_dir)
            remote_projects = api.get_repo_names()

            specified_remotes = remote_projects if all_ else set(projects)

            for project in projects:
                if project not in remote_projects:
                    not_found = True
                    not_found_project = project
                    break
                break

        if not_found:
            # If we don't break out of the with block
            # the spinner will go forever
            msg.warn(
                f"{not_found_project!r} not found on GitHub. Was it a typo?",
                spaced=True,
                exits=1,
            )

        diff = specified_remotes.difference(local_projects)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        pull_diff(
            diff=diff, remotes=remote_projects, force=force, config=config, git=Git()
        )


def pull_diff(
    diff: Set[str], remotes: Set[str], force: bool, config: Config, git: Git
) -> None:
    """
    Helper function to help the pull CLI command with repeated logic.

    Args:
        diff (Set[str]): The difference to clone.
        remotes (Set[str]): Set of remote projects.
        force (bool): The value for the '--force' flag.
        config (Config): The Config object.
    """

    if not diff:
        msg.info("Your local and remote projects are in sync.")
        msg.good("Nothing to pull!", exits=0)

    if not remotes:
        msg.warn("You don't have any remote projects to pull!", spaced=True, exits=1)

    n_diff = len(diff)

    if not force:
        if n_diff <= 3:
            # A managable length to display each one
            typer.confirm(
                f"This will pull down {', '.join(diff)}. Are you sure?", abort=True
            )
        else:
            # Too many to show to look nice, just show number
            typer.confirm(
                f"This will pull down {n_diff} projects. Are you sure?", abort=True
            )

    # If we get here, we're good to go

    if n_diff < 2:
        # Doesn't make sense to spin up a thread pool for 1 repo
        _pull_diff_synchronously(diff=diff, config=config, git=Git())
    else:
        # Let's go concurrency baby!
        _pull_diff_concurrently(diff=diff, config=config, git=Git())


def _pull_diff_synchronously(diff: Set[str], config: Config, git: Git) -> None:
    """
    Helper function to pull the diff synchronously, i.e.
    one repo at a time.

    To be used when the user only wants 1 repo, as it doesn't make sense
    to spin up a thread pool for 1.
    """
    for project in diff:
        # Make a Repo object for each
        repo = Repo(
            owner=config.username,
            name=project,
            local_path=config.projects_dir.joinpath(project),
        )

        msg.info(f"Cloning {project!r}", spaced=True)
        # Because we're cloning just one, makes sense to show clone output
        git.clone(url=repo.clone_url, cwd=config.projects_dir, silent=False)

    msg.good("Done!", spaced=True)


def _pull_diff_concurrently(diff: Set[str], config: Config, git: Git) -> None:
    """
    Helper function to pull the diff concurrently using a pool
    of worker threads.

    Useful when more than 1 repo as the cost of spinning up a thread
    pool is easily outweighed by the average repo clone time.
    """
    # Create a task queue
    to_clone: List[Repo] = []
    for project in diff:
        # Make a Repo object for each to access the attributes
        repo = Repo(
            owner=config.username,
            name=project,
            local_path=config.projects_dir.joinpath(project),
        )
        to_clone.append(repo)

    with concurrent.futures.ThreadPoolExecutor() as executor:

        # Start the cloning and mark each future with it's clone url
        # Set silent=True in git.clone to prevent weird output ordering
        future_to_repo = {
            executor.submit(
                git.clone, url=repo.clone_url, cwd=config.projects_dir, silent=True
            ): repo
            for repo in to_clone
        }

        for future in concurrent.futures.as_completed(future_to_repo):

            current = future_to_repo.get(future)
            if not current:
                # Hopefully absolutely no chance of this happening but I like
                # to always handle the None case from .get
                raise ValueError(f"Repo in task queue for future: {future!r} was None")

            try:
                # Gather the result
                future.result()
            except Exception as err:
                # This should only be very odd things
                msg.fail(f"Error cloning {current.name!r}", text=f"{err}")
            else:
                msg.good(f"Cloned {current.name!r}")



"""
Module responsible for handling poetry environments.

Here we take advantage of poetry's new `local` config setting
to enforce the virtual environment being in the project
without altering the user's base config.

Author: Tom Fleet
Created: 13/07/2021
"""

import shutil
import subprocess
from pathlib import Path
from typing import List, Optional

from pytoil.environments import Environment
from pytoil.exceptions import PoetryNotInstalledError


class PoetryEnv(Environment):
    """
    Poetry EnvManager class.
    """

    def __init__(self, project_path: Path) -> None:
        """
        Representation of a poetry managed virtual environment.

        Most of the work here is just passed through to
        poetry via a subprocess.

        Args:
            project_path (Path): The root path to the current project.
        """
        self._project_path = project_path.resolve()

    def __repr__(self) -> str:
        return self.__class__.__qualname__ + f"(project_path={self.project_path!r})"

    @property
    def project_path(self) -> Path:
        return self._project_path

    @property
    def executable(self) -> Path:
        return self._project_path.joinpath(".venv/bin/python")

    @property
    def info_name(self) -> str:
        return "poetry"

    def raise_for_poetry(self) -> None:
        """
        Helper method to check for the existence of poetry.

        Basically everything in this class relies on poetry being
        installed to function correctly that's why we need this
        method.

        Raises:
            PoetryNotInstalledError: If poetry not installed
        """
        if not bool(shutil.which("poetry")):
            raise PoetryNotInstalledError(
                "Poetry not installed. Cannot install poetry based project."
            )

    def enforce_local_config(self) -> None:
        """
        Ensures any changes to poetry's config such
        as storing the virtual environment in the project
        directory as we do here, do not propegate to the
        user's global poetry config.
        """
        self.raise_for_poetry()

        subprocess.run(
            ["poetry", "config", "virtualenvs.in-project", "true", "--local"],
            check=True,
            cwd=self.project_path,
        )

    def exists(self) -> bool:
        """
        Checks whether the virtual environment exists by a proxy
        check if the `executable` exists.

        If this executable exists then both the project and virtual environment
        must also exist and must be valid.
        """
        return self.executable.exists()

    def create(self, packages: Optional[List[str]] = None) -> None:
        """
        This method is not implemented for poetry environments.

        Use install instead as with poetry, creation and installation
        are handled together.
        """
        raise NotImplementedError

    def install(self, packages: List[str]) -> None:
        """
        Calls `poetry add` under the hood to install
        packages into the environment.

        Args:
            packages (List[str]): List of packages to install.
        """
        self.enforce_local_config()
        subprocess.run(
            ["poetry", "add", *packages, "--quiet"], check=True, cwd=self.project_path
        )

    def install_self(self) -> None:
        """
        Calls `poetry install` under the hood to install the
        current package and all it's dependencies.
        """
        self.enforce_local_config()
        subprocess.run(
            ["poetry", "install", "--quiet"],
            check=True,
            cwd=self.project_path,
            capture_output=True,
        )



from enum import Enum

from pytoil.environments.abstract import Environment
from pytoil.environments.conda import Conda
from pytoil.environments.flit import FlitEnv
from pytoil.environments.poetry import PoetryEnv
from pytoil.environments.reqs import ReqTxtEnv
from pytoil.environments.virtualenv import Venv

__all__ = [
    "Environment",
    "Venv",
    "Conda",
    "FlitEnv",
    "PoetryEnv",
    "ReqTxtEnv",
]


# Choice of virtual environments for a new project
class VirtualEnv(str, Enum):
    """
    Choice of virtualenvs to create in a new project.
    """

    venv = "venv"
    conda = "conda"
    none = "none"
